import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;
import java.util.*;


/**
 * A remote tag server for the GeoVTag application. 
 * 
 * @author Michel Deriaz
 */
public class GeoVTag extends HttpServlet {
  private final String TAG_FILENAME = "tags.bin";  
  private final double NI = Double.NEGATIVE_INFINITY;
  
  private static Object lock = new Object(); 
  private static File tagFile; 
  private Vector<Tag> v;

  
  public void init() {
    tagFile = new File(getServletConfig().getServletContext().getRealPath("/") + TAG_FILENAME);
    synchronized (lock) {
      v = (Vector<Tag>)readObject(tagFile);
      if (v == null) v = new Vector<Tag>();
    }
  }
   
    
  public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
    String latitude = request.getParameter("lat");
    String longitude = request.getParameter("lon");
    String content = fromHexa(request.getParameter("content"));
    int radius = 20; // default value
    try {
      radius = Integer.parseInt(request.getParameter("radius"));
    }
    catch (NumberFormatException e) {}
    if (latitude == null || latitude.equals("") || longitude == null || longitude.equals("")) return;
    String output = "";
    if (content == null) {
      for (int i = 0; i < v.size(); i++) {
        Tag tag = v.get(i);
        double distance = haversine(latitude, longitude, tag.getLat(), tag.getLon());
        if (distance != NI && distance < radius) output += tag + "\n";
      }
      if (output.endsWith("\n")) output = output.substring(0, output.length()-1);
    }
    else {
      v.add(new Tag(latitude, longitude, content));
      synchronized (lock) {
        writeObject(tagFile, v);
      }
      output = "OK";
    }
    response.setContentType("text/plain");
    response.setContentLength(output.length());
    PrintWriter pw = response.getWriter();
    pw.println(output); 
  }


  /**
   * Computes the distance between two points on Earth.
   * @param lat1 latitude of the first point in decimal degrees
   * @param lon1 longitude of the first point in decimal degrees
   * @param lat2 latitude of the second point in decimal degrees
   * @param lon2 longitude of the second point in decimal degrees
   * @return the distance, in meters, between the two points, or MobiVTag.NI if 
   * an error occurs.
   */
  /*
   * Implemented from the formulas found at 
   * http://www.cs.nyu.edu/visual/home/proj/tiger/gisfaq.html
   */
  private double haversine(String lat1, String lon1, String lat2, String lon2) {
    try {
      double lat1Rad = Math.toRadians(Double.parseDouble(lat1));
      double lon1Rad = Math.toRadians(Double.parseDouble(lon1));
      double lat2Rad = Math.toRadians(Double.parseDouble(lat2));
      double lon2Rad = Math.toRadians(Double.parseDouble(lon2));
      double dlat = lat2Rad - lat1Rad;
      double dlon = lon2Rad - lon1Rad;
      double a = Math.sin(dlat/2.0) * Math.sin(dlat/2.0) + Math.cos(lat1Rad) * Math.cos(lat2Rad) * Math.sin(dlon/2.0) * Math.sin(dlon/2.0); 
      double c = 2.0 * Math.asin(Math.min(1.0,Math.sqrt(a)));
      return 6367000.0 * c; // 6367000 = radius of Earth at 45 of latitude 
    }
    catch (Exception e) {
      return NI;
    }
  }

  
  private String fromHexa(String input) {
    if (input == null) return null;
    char[] hexa = input.toCharArray();
    int length = hexa.length / 2;
    byte[] b = new byte[length];
    for (int i = 0; i < length; i++) { 
      int value = (Character.digit(hexa[2*i], 16) << 4) | Character.digit(hexa[2*i+1], 16);
      if (value > 127) value -= 256;
      b[i] = (byte)value;
    } 
    return new String(b);
  } 
  
  
  private void writeObject(File f, Object object) {
    try {
      FileOutputStream fos = new FileOutputStream(f);
      ObjectOutputStream oos = new ObjectOutputStream(fos);
      oos.writeObject(object);
      oos.flush();
      fos.close();
      oos.close();
    }
    catch (IOException e) {
      System.out.println(e);
    }
  }
  

  private Object readObject(File f) {
    Object obj = null;
    try {
      FileInputStream fis = new FileInputStream(f);
      ObjectInputStream ois = new ObjectInputStream(fis);
      obj = ois.readObject();
      fis.close();
      ois.close();
    }
    catch (IOException e) { // file doesn't exist
    }
    catch (ClassNotFoundException e) { 
      System.out.println(e);
    }
    return obj;
  }    
}